# -*- coding: utf-8 -*-
"""Widgets."""

from .history import OQHistoryWidget, OQHistoryWidgetDefaultHeader
from .list import OQWidgetList, OQWidgetListDefaultHeader
from .widget import OQWidget

__all__ = [
    "OQWidget",
    "OQWidgetListDefaultHeader",
    "OQWidgetList",
    "OQHistoryWidgetDefaultHeader",
    "OQHistoryWidget",
]
